#!/usr/bin/perl 
# 2crd.pl
# convert textfile to a hollerith coded punched card image file
# lower case characters are converted to upper case
# tabs are expanded to the specified tab-stops
#
# input: max 80 characters excl. <CR><LF>; if more then truncate.
#
# Output (with) argument h: fixed length record format 80 16 bit words, where 
# the 12 least significant bits of a word are used for a 12 bit hollerith code. 
# Only the 64 ASCII character set is used and translated to hollerith code. 
# A character not within that set is translated to <space>
#
# With the argument a, the hollerith conversion is skipped. The ascii code
# is written as 16 bits directly into the card image (ascii in the lower
# byte and 0 to the ms byte)
# 
# Either with hollerith code or ascii code, the access to the card reader
# is similar: each read access delivers 16 bits. 
#
##  usage: 2cdr.pl [a|h] text.txt|text.src
## output: text.crd
#
use strict;
use bytes;
my $DEBUG = 0;

# .src source input fields (P800 assembler)
##########################
# the source input record is 80 characters maximum (punched card image) and 
# divided in 4 fields: label, operation, operand, comment
#
# default input record layout:
#***** !**********!!**************!!****!**************************************!
#label operation   operand         comment
#                                       comment
# record layout can be changed by adapting the following settings.
#
my $tab1 = 0;               # label field
my $fl1  = 6;               # +1
my $tab2 = 7;               # operation field
my $fl2  = 12;
my $tab3 = 19;              # operand field
# standard operand field
#my $fl3  = 16; 
#my $tab4 = 35;              # comment field
#my $fl4  = 45;
# larger operand field
my $fl3  = 21; 
my $tab4 = 40;              # comment field
my $fl4  = 40;

my $f1;                     # input field1 char 0-5 (l=6)
my $f2;                     # input field2 char 7-18 (l=12)
my $f3;                     # input field3 char 19-34 (l=16)
my $f4;                     # input field4 char 35-79 (l=45)
my $ch;						# input character
my $line = "";				# expanded input record
my $cin  = 0;				# input record char counter
my $cout = 0;				# expanded record char counter
my $quot = 0;               # in quoted string flag
my $ll;						# input record length
my $l;                      # temp
my $i;                      # temp

my $mode = shift;

if($mode =~ /a/)
{
	print "Ascii conversion\n";
}
elsif($mode =~ /h/)
{
	print "Hollerith conversion\n";
}
else
{
	print "Code conversion argument missing\n";
	exit;
}

my $fi = shift (@ARGV);				# source file
if ($fi eq "" &&  $fi !~ /\.txt/ && $fi !~ /\.src/)
{
    print "No filename, or wrong .txt or .src extention $fi\n";
    exit;
}

my $fn = "";                        # file name
my $ty = "";                        # file type
my $fo = "";						# output file (card image)

# If the extension is .src, tabs are expanded first.
# The result is a file with expanded records with a .txt extension
# In case the provided input file has already an .txt extension, tab expansion
# is skipped and record conversion is entered directly. 
# In case an input file with .txt extension is provided which also is a P800
# source file for the assembler, then it is expected to have the fields 
# formatted in the right colums for the assembler.

if( $fi =~ /\.src/)
{
	($fn,$ty) = split(/\./, $fi);
	$fo = $fn . ".txt";
	print "Expand tabs in $fi to file $fo\n";
	print "Start reading $fi \n";
	open (IN,$fi) || die "cannot open inputfile $fi: $!";
	open (OUT,">$fo") || die "cannot open inputfile $fo: $!";
	while(<IN>)
	{
		$line = "";
		$cin  = 0;
		$cout = 0;
		$quot = 0;
		chomp;
		$ll = length($_);
		# expand tabs
		if(($ll > 0) && (substr($_,$cin,1) ne "*"))
		{
			while($cin < $ll)   
			{
				$ch=substr($_,$cin,1);			        # get next char from input record
				$cin=$cin+1;
				if( $ch eq "\'" && $quot == 0 )
				{
					$quot = 1;							# set the in-quoted-string-flag
				}
				elsif( $ch eq "\'" && $quot == 1 )
				{
					$quot = 0;							# reset the in-quoted-string-flag
				}
				if($quot)
				{
					# in-quoted-string: don't expand tabs
					$line=$line . $ch;
					$cout = $cout + 1;
				}
				else
				{
                    if( $ch =~ /\\/ || $ch =~ /\ /)		                # replace tab by one or more spaces						
#					if( $ch =~ /\t/ || $ch =~ /\\/ || $ch =~ /\ /)		# replace tab by one or more spaces
#					if( $ch =~ /\t/ || $ch =~ /\\/)		                # replace tab by one or more spaces
					{
						# replace tab by space
						$line=$line . " ";					# expanded input record
						$cout=$cout+1;
						if($cout <= $tab2)
						{
							while($cout<$tab2)
							{
									$line=$line . " ";	# expanded input record
					                $cout=$cout+1;
							}
							next;
						}
						if($cout <= $tab3)
						{
							while($cout<$tab3)
							{
									$line=$line . " ";	# expanded input record
					                $cout=$cout+1;
							}
							next;
						}
						if($cout <= $tab4)
						{
							while($cout<$tab4)
							{
									$line=$line . " ";	# expanded input record
					                $cout=$cout+1;
							}
							next;
						}
					}
					else
					{
						$line=$line . $ch;
						$cout = $cout + 1;
					}
				}
			}
		}
		else
		{
			$line = $_;		# comment line
		}
		print OUT $line . "\n";
	}
	close (IN)   || die "can't close $fi: $!";
	close (OUT)  || die "can't close $fo: $!";	
	print "$fi expanded into $fo\n";
	$fi = $fo;
}

($fn,$ty) = split(/\./, $fi);
$fo = $fn . ".crd";

print "Convert $fi to the card image file $fo\n";
print "Start reading $fi\n";

open (IN,$fi) || die "cannot open inputfile $fi: $!";
open (BIN, ">:raw", $fo) || die "cannot open outputfile $fo: $!";

my $cin=0;		# number of lines read in
my $cout=0;		# number of lines punched
my $l;			# input record length
my $ch;			# char input record
my $ptrin=0;	# input record char pointer 0..n-1
my $ptrout=0;	# output record char pointer 0..m-1
my $line = "";	# assembled expanded input text record 

while (<IN>)
{   
	$cin=$cin+1;
	$ptrin=0;
    $ptrout=0;	
	$line = "";
	$l=length($_);
	tr/[a-z]/[A-Z]/;			# to upper case
 	while($ptrin < $l && $ptrin < 80)   
    {
        $ch=substr($_,$ptrin,1);			# get next char from input record

		next if($ch eq "\r");                       # skip
		if($ch ne "\n")
		{
			$line=$line . $ch;						#.text file				
			$ch=unpack("C",$ch) ;
			print chr($ch);							# console
			crdpunch($ch);							#.crd file
    		$ptrout=$ptrout+1;
		}
		else
		{
			last;
		}
		$ptrin=$ptrin+1;			
	}
	#
	# end of input record or 80 characters from input record
	#
	$i=$ptrin;
    while($ch ne "\n")                  # skip to record end
	{
		# truncate remaining part of record
		$ch=substr($_,$i,1);			# get next char from input record
		$i=$i+1;
	}
	while($ptrin < 80)                  # punch spaces until eor
	{
		crdpunch(0x20);
	    $ptrin=$ptrin+1;
	}
	print "\n";											#console
	$line = $line . "\n";
	$cout=$cout+1;
}

close (IN)  || die "can't close $fi: $!";
close (BIN) || die "can't close $fo: $!";

print "End of card image conversion conversion\n";
print "$cin  lines read from $fi\n";
print "$cout lines writtten to $fo\n";

# punch ascii char on card punch in hollerith code
sub crdpunch {
	# ascii => hollerith conversion
	my @ah;
	$ah[0x20] = 0x000; # space
	$ah[0x21] = 0x806; # !
	$ah[0x22] = 0x006; # "
	$ah[0x23] = 0x042; # #
	$ah[0x24] = 0x442; # $
	$ah[0x25] = 0x222; # %
	$ah[0x26] = 0x800; # &
	$ah[0x27] = 0x012; # '
	$ah[0x28] = 0x812; # (
	$ah[0x29] = 0x412; # )
	$ah[0x2A] = 0x422; # *
	$ah[0x2B] = 0x80A; # +
	$ah[0x2C] = 0x242; # ,
	$ah[0x2D] = 0x400; # -
	$ah[0x2E] = 0x842; # .
	$ah[0x2F] = 0x300; # /
	
	$ah[0x30] = 0x200; # 0
	$ah[0x31] = 0x100; # 1
	$ah[0x32] = 0x080; # 2
	$ah[0x33] = 0x040; # 3
	$ah[0x34] = 0x020; # 4
	$ah[0x35] = 0x010; # 5
	$ah[0x36] = 0x008; # 6
	$ah[0x37] = 0x004; # 7
	$ah[0x38] = 0x002; # 8
	$ah[0x39] = 0x001; # 9
	$ah[0x3A] = 0x082; # :
	$ah[0x3B] = 0x40A; # ;
	$ah[0x3C] = 0x822; # <
	$ah[0x3D] = 0x00A; # =
	$ah[0x3E] = 0x20A; # >
	$ah[0x3F] = 0x206; # ?	
	
	$ah[0x40] = 0x022; # @
	$ah[0x41] = 0x900; # A
	$ah[0x42] = 0x880; # B
	$ah[0x43] = 0x840; # C
	$ah[0x44] = 0x820; # D
	$ah[0x45] = 0x810; # E
	$ah[0x46] = 0x808; # F
	$ah[0x47] = 0x804; # G
	$ah[0x48] = 0x802; # H
	$ah[0x49] = 0x801; # I
	$ah[0x4A] = 0x500; # J
	$ah[0x4B] = 0x480; # K
	$ah[0x4C] = 0x440; # L
	$ah[0x4D] = 0x420; # M
	$ah[0x4E] = 0x410; # N
	$ah[0x4F] = 0x408; # O		
	
	$ah[0x50] = 0x404; # P
	$ah[0x51] = 0x402; # Q
	$ah[0x52] = 0x401; # R
	$ah[0x53] = 0x280; # S
	$ah[0x54] = 0x240; # T
	$ah[0x55] = 0x220; # U
	$ah[0x56] = 0x210; # V
	$ah[0x57] = 0x208; # W
	$ah[0x58] = 0x204; # X
	$ah[0x59] = 0x202; # Y
	$ah[0x5A] = 0x201; # Z
	$ah[0x5B] = 0x882; # [
	$ah[0x5C] = 0x282; # \
	$ah[0x5D] = 0x482; # ]
	$ah[0x5E] = 0x406; # ^
	$ah[0x5F] = 0x212; # _
	
	
    my($a) = shift;
	# print "A" . prhex(32,$a) . " ";	
    my $byte;
	if($a < 0x20 || $a > 0x5F) # within 64 ASCII char set ?
	{
		$a = 0x20; # blank
	}
	######## CONVERSION #########
	if($mode =~ /h/)
	{
		$a=$ah[$a];      # ascii => hollerith
	}
	#############################
	# print "H" . prhex(32,$a) . " ";
    $byte = pack("C", (($a & 0xFF00) >> 8)); # ms byte
    print BIN $byte;
    $byte = pack("C", $a & 0xFF);            # ls byte
    print BIN $byte;	
}

sub prhex {
	# print hex
	# $n is number of bits
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	if($n == 0)
	{
		return "";
	}
	elsif($n<5)
	{
		$s = sprintf("%01X", $v);	
	}
	elsif($n<9)
	{
		$s = sprintf("%02X", $v);
	}
	elsif($n<13)
	{		
		$s = sprintf("%03X", $v);
	}
	elsif($n<17)
	{		
		$s = sprintf("%04X", $v);
	}
	elsif($n<21)
	{
		$s = sprintf("%05X", $v);
	}
	elsif($n<25)
	{						
		$s = sprintf("%06X", $v);
	}
	elsif($n<29)
	{		
		$s = sprintf("%07X", $v);
	}
	elsif($n<33)
	{
		$s = sprintf("%08X", $v);
	}
	# print "$s ";
	return $s;
}

